"""builder for all the representations in VRE. Can be used outside of VRE too, see examples/notebooks."""
from typing import Type
import imp # pylint: disable=deprecated-module
from omegaconf import DictConfig, OmegaConf

from ..logger import vre_logger as logger
from ..utils import topological_sort
from .representation import Representation
from .learned_representation_mixin import LearnedRepresentationMixin
from .compute_representation_mixin import ComputeRepresentationMixin
from .io_representation_mixin import IORepresentationMixin

# pylint: disable=import-outside-toplevel, redefined-builtin, too-many-branches, too-many-statements, cyclic-import
def build_representation_type(repr_type: str) -> Type[Representation]:
    """Gets the representation type from a type and a name (the two identifiers of a representation)"""
    obj_type: Type[Representation] | None = None
    assert len(repr_type.split("/")) == 2, f"No extra '/' allowed in the type. Got: '{repr_type}'. Correct: depth/dpt"

    if repr_type == "default/rgb":
        from .color.rgb import RGB
        obj_type = RGB
    elif repr_type == "default/hsv":
        from .color.hsv import HSV
        obj_type = HSV
    elif repr_type == "soft-segmentation/python-halftone":
        from .soft_segmentation.halftone import Halftone
        obj_type = Halftone
    elif repr_type == "soft-segmentation/generalized_boundaries":
        from .soft_segmentation.generalized_boundaries import GeneralizedBoundaries
        obj_type = GeneralizedBoundaries
    elif repr_type == "soft-segmentation/fastsam":
        from .soft_segmentation.fastsam import FastSam
        obj_type = FastSam
    elif repr_type == "edges/dexined":
        from .edges.dexined import DexiNed
        obj_type = DexiNed
    elif repr_type == "edges/canny":
        from .edges.canny import Canny
        obj_type = Canny
    elif repr_type == "depth/dpt":
        from .depth.dpt import DepthDpt
        obj_type = DepthDpt
    elif repr_type == "depth/marigold":
        from .depth.marigold import Marigold
        obj_type = Marigold
    elif repr_type == "optical-flow/rife":
        from .optical_flow.rife import FlowRife
        obj_type = FlowRife
    elif repr_type == "optical-flow/raft":
        from .optical_flow.raft import FlowRaft
        obj_type = FlowRaft
    elif repr_type == "semantic-segmentation/safeuav":
        from .semantic_segmentation.safeuav import SafeUAV
        obj_type = SafeUAV
    elif repr_type == "semantic-segmentation/mask2former":
        from .semantic_segmentation.mask2former import Mask2Former
        obj_type = Mask2Former
    elif repr_type == "normals/depth-svd":
        from .normals.depth_svd import DepthNormalsSVD
        obj_type = DepthNormalsSVD
    else:
        raise ValueError(f"Unknown type: '{repr_type}'")
    return obj_type

def _validate_repr_cfg(repr_cfg: dict, name: str):
    assert isinstance(repr_cfg, dict), f"Broken format (not a dict) for {name}. Type: {type(repr_cfg)}."
    valid_keys = {"type", "parameters", "dependencies", "compute_parameters", "learned_parameters", "io_parameters"}
    required_keys = {"type", "parameters", "dependencies"}
    assert set(repr_cfg).issubset(valid_keys), f"{name} wrong keys: {repr_cfg.keys()}"
    assert (diff := required_keys.difference(repr_cfg)) == set(), f"{name} missing keys: {diff}"
    assert isinstance(repr_cfg["parameters"], dict), type(repr_cfg["parameters"])
    assert name.find("/") == -1, "No '/' allowed in the representation name. Got '{name}'"

def build_representation_from_cfg(repr_cfg: dict, name: str, built_so_far: dict[str, Representation],
                                  compute_representations_defaults: dict | None = None,
                                  learned_representations_defaults: dict | None = None,
                                  io_representations_defaults: dict | None = None) -> Representation:
    """Builds a representation given a dict config and a name."""
    _validate_repr_cfg(repr_cfg, name)
    logger.info(f"Building '{repr_cfg['type']}' (vre name: {name})")
    obj_type = build_representation_type(repr_cfg["type"])
    dependencies = [built_so_far[dep] for dep in repr_cfg["dependencies"]]
    obj: Representation = obj_type(name=name, dependencies=dependencies, **repr_cfg["parameters"])

    if isinstance(obj, LearnedRepresentationMixin):
        defaults = {} if learned_representations_defaults is None else learned_representations_defaults
        defaults = OmegaConf.to_container(defaults, resolve=True) if isinstance(defaults, DictConfig) else defaults
        if "learned_parameters" in repr_cfg:
            repr_learned_parameters = {**defaults, **repr_cfg.get("learned_parameters", {})}
            logger.debug(f"[{obj}] Setting node specific 'Learned' params: {repr_learned_parameters}")
        else:
            repr_learned_parameters = defaults
            logger.debug(f"[{obj}] Setting default 'Learned' params: {repr_learned_parameters}")
        obj.set_learned_params(**repr_learned_parameters)
    else:
        assert "learned_parameters" not in repr_cfg, f"Learned parameters not allowed for {name}"

    if isinstance(obj, ComputeRepresentationMixin):
        defaults = {} if compute_representations_defaults is None else compute_representations_defaults
        defaults = OmegaConf.to_container(defaults, resolve=True) if isinstance(defaults, DictConfig) else defaults
        if "compute_parameters" in repr_cfg:
            repr_compute_params = {**defaults, **repr_cfg.get("compute_parameters", {})}
            logger.debug(f"[{obj}] Setting node specific 'Compute' params: {repr_compute_params}")
        else:
            repr_compute_params = defaults
            logger.debug(f"[{obj}] Setting default 'Compute' params: {repr_compute_params}")
        obj.set_compute_params(**repr_compute_params)
    else:
        assert "compute_parameters" not in repr_cfg, f"Compute parameters not allowed for {name}"

    if isinstance(obj, IORepresentationMixin):
        defaults = {} if io_representations_defaults is None else io_representations_defaults
        defaults = OmegaConf.to_container(defaults, resolve=True) if isinstance(defaults, DictConfig) else defaults
        if "io_parameters" in repr_cfg:
            repr_io_params = {**defaults, **repr_cfg.get("io_parameters", {})}
            logger.debug(f"[{obj}] Setting node specific 'I/O' params: {repr_io_params}")
        else:
            repr_io_params = defaults
            logger.debug(f"[{obj}] Setting default 'I/O' params: {repr_io_params}")
        obj.set_io_params(**repr_io_params)
    else:
        assert "io_parameters" not in repr_cfg, f"I/O parameters not allowed for {name}"
    return obj

def build_representations_from_cfg(cfg: DictConfig | dict) -> dict[str, Representation]:
    """builds a dict of representations given a dict config (yaml file)"""
    cfg: dict = OmegaConf.to_container(cfg, resolve=True) if isinstance(cfg, DictConfig) else cfg
    assert len(repr_cfg := cfg["representations"]) > 0 and isinstance(repr_cfg, dict), repr_cfg

    tsr: dict[str, Representation] = {}
    logger.debug("Doing topological sort...")
    dep_graph = {}
    for repr_name, repr_cfg_values in repr_cfg.items():
        _validate_repr_cfg(repr_cfg_values, repr_name)
        assert isinstance(repr_cfg_values, dict), f"{repr_name} not a dict cfg: {type(repr_cfg_values)}"
        dep_graph[repr_name] = repr_cfg_values["dependencies"]
    topo_sorted = {k: repr_cfg[k] for k in topological_sort(dep_graph)}

    for name, repr_cfg in topo_sorted.items():
        obj = build_representation_from_cfg(repr_cfg, name, tsr, cfg.get("default_compute_parameters"),
                                            cfg.get("default_learned_parameters"), cfg.get("default_io_parameters"))
        tsr[name] = obj
    return tsr

def add_external_representations(representations: dict[str, Representation], external_path: str,
                                 cfg: DictConfig) -> dict[str, Representation]:
    """adds external representations from an provided path in the format: /path/to/script.py:fn_name"""
    path, fn = external_path.split(":")
    external_representations: dict[str, Representation] = getattr(imp.load_source("external", path), fn)()
    assert all(isinstance(v, IORepresentationMixin) for v in external_representations.values())
    assert all(isinstance(v, ComputeRepresentationMixin) for v in external_representations.values())
    assert all(isinstance(v, Representation) for v in external_representations.values())
    assert (clash := set(external_representations.keys()).intersection(representations)) == set(), clash
    logger.info(f"Adding {list(external_representations)} from {path}")
    for repr in external_representations.values():
        repr.set_compute_params(**cfg.get("default_compute_parameters", {}))
        repr.set_io_params(**cfg.get("default_io_parameters", {}))
        if isinstance(repr, LearnedRepresentationMixin):
            repr.set_learned_parameters(**cfg.get("default_learned_parameters", {}))
    representations = {**representations, **external_representations}
    tsr = topological_sort({r.name: [_r.name for _r in r.dependencies] for r in representations.values()})
    representations = {k: representations[k] for k in tsr}
    return representations



"""Init file"""
# pylint: disable=reimported, wrong-import-position
import warnings
from lovely_tensors import monkey_patch
from diffusers.utils.logging import disable_progress_bar
import matplotlib
try:
    matplotlib.use("TkAgg") # sometimes we default to PyQt5 if this is not set here which crashes randomly....
except ImportError:
    matplotlib.use("Agg")

from .video_representations_extractor import VideoRepresentationsExtractor, VideoRepresentationsExtractor as VRE
from .representations import Representation, ReprOut
from .utils.vre_video import FFmpegVideo

try:
    import pdbp
except ImportError:
    pass

monkey_patch()
disable_progress_bar()

warnings.filterwarnings("ignore", "You are using `torch.load` with `weights_only=False`*.")
warnings.filterwarnings("ignore", "`clean_up_tokenization_spaces` was not set*.")



"""Depth normals representation using SVD."""
from overrides import overrides
import numpy as np

from vre.utils import VREVideo
from vre.representations import Representation, ReprOut, ComputeRepresentationMixin, NpIORepresentation
from vre.representations.normals.depth_svd.depth_svd_impl import (
    fov_diag_to_intrinsic, get_sampling_grid, get_normalized_coords, depth_to_normals)

class DepthNormalsSVD(Representation, ComputeRepresentationMixin, NpIORepresentation):
    """
    General method for estimating normals from a depth map (+ intrinsics): a 2D window centered on each pixel is
    projected into 3D and then a plane is fitted on the 3D pointcloud using SVD.
    """
    def __init__(self, sensor_fov: int, sensor_width: int, sensor_height: int, window_size: int,
                 input_downsample_step: int = None, stride: int = None, max_distance: float = None,
                 min_valid_count: int = None, **kwargs):
        Representation.__init__(self, **kwargs)
        ComputeRepresentationMixin.__init__(self)
        NpIORepresentation.__init__(self)
        assert window_size % 2 == 1, f"Expected odd window size. Got: {window_size}"
        self.sensor_fov = sensor_fov
        self.sensor_width = sensor_width
        self.sensor_height = sensor_height
        self.window_size = window_size
        self.stride = stride if stride is not None else 1
        self.input_downsample_step = input_downsample_step if input_downsample_step is not None else 1
        self.max_dist = max_distance if max_distance is not None else -1
        self.min_valid = min_valid_count if min_valid_count is not None else 0
        assert len(self.dependencies) == 1, f"Expected exactly one depth method, got: {self.dependencies}"
        self._grid_cache = {}

    @overrides
    def compute(self, video: VREVideo, ixs: list[int]):
        assert self.data is None, f"[{self}] data must not be computed before calling this"
        assert (A := self.dependencies[0].data) is not None and self.dependencies[0].data.key == ixs, (A.key, ixs)
        depths = self.dependencies[0].data.output
        assert len(depths.shape) == 3, f"Expected (T, H, W) got: {depths.shape}"
        res = np.array([self._make_one_normal(depth) for depth in depths])
        self.data = ReprOut(frames=np.array(video[ixs]), output=res, key=ixs)

    @overrides
    def make_images(self) -> np.ndarray:
        assert self.data is not None, f"[{self}] data must be first computed using compute()"
        return (self.data.output * 255).astype(np.uint8)

    def _make_one_normal(self, depth: np.ndarray) -> np.ndarray:
        # TODO: batch vectorize this if possible
        sampling_grid, normalized_grid = self._get_grid(depth)
        if self.input_downsample_step is not None:
            depth = depth[:: self.input_downsample_step, :: self.input_downsample_step]
        normals = depth_to_normals(depth, sampling_grid, normalized_grid, self.max_dist, self.min_valid)
        normals = (normals.astype(np.float32) + 1) / 2
        return normals

    def _get_grid(self, depth: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        height, width = depth.shape[:2]
        if (height, width) in self._grid_cache:
            return self._grid_cache[(height, width)]
        if self.input_downsample_step is not None:
            depth = depth[:: self.input_downsample_step, :: self.input_downsample_step]
        depth_height, depth_width = depth.shape[:2]
        sampling_grid = get_sampling_grid(depth_width, depth_height, self.window_size, self.stride)
        K = fov_diag_to_intrinsic(self.sensor_fov, (self.sensor_width, self.sensor_height), (depth_width, depth_height))
        normalized_grid = get_normalized_coords(depth_width, depth_height, K)
        self._grid_cache[(height, width)] = sampling_grid, normalized_grid
        return sampling_grid, normalized_grid



"""FakeVideo module"""
from typing import Union
from pathlib import Path
import numpy as np
import ffmpeg

class FakeVideo:
    """
    FakeVideo -- class used to test representations with a given numpy array.
    If 'frames' is provided, then these are the frames indexes, i.e. video[ix] returns self.data[self.frames_ix[ix]]
    This can be used to create a FakeVideo out of a directory with files: [1.png, 10.png, 100.png]
    """
    def __init__(self, data: np.ndarray, fps: float, frames: list[int] | None = None):
        assert len(data) > 0, "No data provided"
        self.data = data
        self.fps = fps
        self.frame_shape = data.shape[1:]
        self.file = f"FakeVideo {self.data.shape}"
        self.frames = list(range(len(data))) if frames is None else frames
        assert 0 < len(self.frames) <= 1_000_000 # max 1M frames to keep it tight
        assert len(self.frames) == len(self.data), (self.frames, self.data)
        assert all(isinstance(frame, int) for frame in self.frames), self.frames
        assert sorted(set(self.frames)) == self.frames, self.frames
        self.frames_ix = dict(zip(self.frames, range(len(self.frames)))) # {ix: frame}

    @property
    def shape(self):
        """returns the shape of the data"""
        return self.data.shape

    def __len__(self) -> int:
        return self.frames[-1] + 1

    def __getitem__(self, ix: int | list[int] | slice) -> np.ndarray:
        if isinstance(ix, list):
            return np.array([self[_ix] for _ix in ix])
        if isinstance(ix, slice):
            return np.array([self[_ix] for _ix in range(ix.start, ix.stop)])
        return self.data[self.frames_ix[ix]]

    def __repr__(self):
        return f"[FakeVideo] FPS: {self.fps}. Len: {len(self.data)}. Frame shape: {self.data.shape[1:]}. FPS: "

class FFmpegVideo:
    """FFmpegVideo -- reads data from a video using ffmpeg"""
    def __init__(self, path: Path, cache_len: int = 30):
        self.path = Path(path)
        assert self.path.exists(), f"Video '{self.path}' doesn't exist"
        self.probe = ffmpeg.probe(self.path)
        self.stream_info = next((stream for stream in self.probe["streams"] if stream["codec_type"] == "video"), None)
        self.fps = eval(self.stream_info["avg_frame_rate"]) # pylint: disable=eval-used
        self.width = int(self.stream_info["width"])
        self.height = int(self.stream_info["height"])
        self.total_frames = int(float(self.stream_info["nb_frames"])) if "nb_frames" in self.stream_info else None
        self.frame_shape = (self.height, self.width, 3)
        self.shape = (self.total_frames, *self.frame_shape)

        self.cache = []
        self.cache_max_len = cache_len
        self.cache_start_frame = None
        self.cache_end_frame = None
        self.video_process = None

    def _start_ffmpeg_process(self, start_time):
        """
        Start an ffmpeg process from the nearest keyframe to the requested start_time.
        This will load all frames from this keyframe to the next keyframe.
        """
        # Close any existing process
        if self.video_process is not None:
            self.video_process.stdout.close()
            self.video_process.stderr.close()
            self.video_process.terminate()

        # Seek to the nearest keyframe before `start_time`
        self.video_process = (
            ffmpeg
            .input(self.path, ss=start_time)
            .output("pipe:", format="rawvideo", pix_fmt="rgb24")
            .run_async(pipe_stdout=True, pipe_stderr=True)
        )

    def _cache_frames(self, start_frame: int):
        """
        Cache all frames between the current keyframe and the next keyframe, starting at start_frame.
        """
        start_time = start_frame / self.fps
        self.cache = []
        self._start_ffmpeg_process(start_time)

        # Read frames until the end of the current keyframe range
        while True:
            in_bytes = self.video_process.stdout.read(self.width * self.height * 3)
            if not in_bytes or len(self.cache) > self.cache_max_len:
                break
            frame = np.frombuffer(in_bytes, np.uint8).reshape([self.height, self.width, 3])
            self.cache.append(frame)

        # Set cache boundaries in terms of frame numbers
        self.cache_start_frame = start_frame
        self.cache_end_frame = start_frame + len(self.cache)

    def get_frame_by_number(self, frame_number: int) -> np.ndarray:
        """Retrieve a frame from the video by frame number, using nearby frames caching."""
        assert isinstance(frame_number, int), type(frame_number)
        assert 0 <= frame_number < self.total_frames, f"Frame out of bounds: {frame_number}. Len: {len(self)}"

        # Load new cache if the requested frame is outside the current cache range
        if self.cache_start_frame is None or not self.cache_start_frame <= frame_number < self.cache_end_frame:
            # keyframe_frame = int(frame_number - (frame_number % (1 / self.fps)))  # Nearest keyframe frame number
            self._cache_frames(frame_number)

        # Calculate the index within the cache
        cache_index = frame_number - self.cache_start_frame
        return self.cache[cache_index]

    def __repr__(self):
        return f"[FFmpegVideo] Path: {self.path}. FPS: {self.fps}. Len: {len(self)}. Frame shape: {self.frame_shape}."

    def __len__(self) -> int:
        return self.total_frames

    def __getitem__(self, ix: int | list[int] | slice) -> np.ndarray:
        if isinstance(ix, list):
            return np.array([self[_ix] for _ix in ix])
        if isinstance(ix, slice):
            return np.array([self[_ix] for _ix in range(ix.start, ix.stop)])
        return self.get_frame_by_number(ix)

    def __del__(self):
        """Clean up the ffmpeg process when done."""
        if hasattr(self, "video_process") and self.video_process is not None: # in case it throws in the constructor
            self.video_process.stdout.close()
            self.video_process.stderr.close()
            self.video_process.terminate()

VREVideo = Union[FFmpegVideo, FakeVideo]
